from typing import Iterable
from biolib.app import BioLibApp
from biolib.biolib_api_client import JobState
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger
from biolib._internal.push_application import DockerStatusUpdate, process_docker_status_updates
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi


def download_container_from_uri(uri: str) -> None:
    docker_client = BiolibDockerClient.get_docker_client()
    app = BioLibApp(uri=uri)
    job_response = BiolibJobApi.create(app_version_id=app.version["public_id"])
    job_uuid = job_response["uuid"]
    auth_config = {'username': 'biolib', 'password': f',{job_uuid}'}
    try:
        docker_image_uri = job_response["app_version"]["modules"][0]["absolute_image_uri"]

        logger.info(msg=f"Pulling Docker image for {uri}...")

        repo, tag = docker_image_uri.split(':')
        pull_status_updates: Iterable[DockerStatusUpdate] = docker_client.api.pull(
            decode=True,
            repository=repo,
            stream=True,
            tag=tag,
            auth_config=auth_config
        )
        process_docker_status_updates(status_updates=pull_status_updates, action='Pulling')
        image = docker_client.images.get(docker_image_uri)
        app_uri_repo, app_uri_tag = app.uri.lower().split(':')
        image.tag(repository=app_uri_repo, tag=app_uri_tag)
        docker_client.images.remove(docker_image_uri, force=True)
        BiolibJobApi.update_state(job_uuid=job_response["public_id"], state=JobState.COMPLETED)

    except Exception as error:
        logger.error(msg=f"Could not pull Docker image for {uri} due to {error}")
        BiolibJobApi.update_state(job_uuid=job_response["public_id"], state=JobState.FAILED)



from biolib.typing_utils import TypedDict, Optional, Literal, List

JobState = Literal['in_progress', 'completed', 'failed', 'cancelled']


class _BaseCloudJobDict(TypedDict):
    created_at: str
    finished_at: Optional[str]
    uuid: str
    error_code: int


class CloudJobDict(_BaseCloudJobDict):
    started_at: Optional[str]
    compute_node_url: Optional[str]


class CloudJobStartedDict(_BaseCloudJobDict):
    started_at: str
    compute_node_url: str


class Result(TypedDict):
    name: str


class JobDict(TypedDict):
    app_uri: str
    auth_token: str
    created_at: str
    ended_at: Optional[str]
    runtime_seconds: int
    main_result: Result
    started_at: str
    state: JobState
    uuid: str
    cloud_job: Optional[CloudJobDict]


class BasePaginatedResponse(TypedDict):
    current_page_number: int
    object_count: int
    page_count: int
    page_size: int


class JobsPaginatedResponse(BasePaginatedResponse):
    results: List[JobDict]



from biolib._internal.push_application import push_application as _push_application
from biolib._internal.push_application import set_app_version_as_active as _set_app_version_as_active

from biolib.app import BioLibApp as _BioLibApp

def push_app_version(uri: str, path: str) -> _BioLibApp:
    push_data = _push_application(
        app_uri=uri,
        app_path=path,
        app_version_to_copy_images_from=None,
        is_dev_version=True)
    uri = f'{push_data["app_uri"]}:{push_data["sematic_version"]}'
    return _BioLibApp(uri)

def set_app_version_as_default(app_version: _BioLibApp) -> None:
    app_version_uuid = app_version.version['public_id']
    _set_app_version_as_active(app_version_uuid)



from datetime import datetime, timedelta
# from urllib.parse import urlparse, parse_qs

from biolib.biolib_logging import logger

from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_binary_format.utils import RemoteEndpoint


class RemoteJobStorageResultEndpoint(RemoteEndpoint):
    def __init__(self, job_id: str, job_auth_token: str):
        self._job_id = job_id
        self._job_auth_token = job_auth_token
        self._expires_at = None
        self._presigned_url = None

    def get_remote_url(self):
        if not self._presigned_url or datetime.utcnow() > self._expires_at:
            self._presigned_url = BiolibJobApi.get_job_storage_download_url(
                job_auth_token=self._job_auth_token,
                job_uuid=self._job_id,
                storage_type='results'
            )
            self._expires_at = datetime.utcnow() + timedelta(minutes=8)
            # TODO: Use expires at from url
            # parsed_url = urlparse(self._presigned_url)
            # query_params = parse_qs(parsed_url.query)
            # time_at_generation = datetime.datetime.strptime(query_params['X-Amz-Date'][0], '%Y%m%dT%H%M%SZ')
            # self._expires_at = time_at_generation + timedelta(seconds=int(query_params['X-Amz-Expires'][0]))
            logger.debug(f'Job "{self._job_id}" fetched presigned URL with expiry at {self._expires_at.isoformat()}')

        return self._presigned_url



import logging
import sys

import click

import biolib.lfs
from biolib import biolib_errors
from biolib.biolib_logging import logger_no_user_data, logger
from biolib.lfs import push_large_file_system, create_large_file_system, describe_large_file_system, prune_lfs_cache
from biolib.typing_utils import Optional


@click.group(help='Manage Large File Systems')
def lfs() -> None:
    pass


@lfs.command(help='Create a Large File System')
@click.argument('uri', required=True)
def create(uri: str) -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)
    create_large_file_system(lfs_uri=uri)


@lfs.command(help='Push a new version of a Large File System')
@click.argument('uri', required=True)
@click.option('--path', required=True, type=click.Path(exists=True))
@click.option('--chunk-size', default=None, required=False, type=click.INT, help='The size of each chunk (In MB)')
def push(uri: str, path: str, chunk_size: Optional[int]) -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)
    try:
        push_large_file_system(lfs_uri=uri, input_dir=path, chunk_size_in_mb=chunk_size)
    except biolib_errors.BioLibError as error:
        print(f'An error occurred:\n{error.message}', file=sys.stderr)
        exit(1)


@lfs.command(help='Download a file from a Large File System')
@click.argument('uri', required=True)
@click.option('--file-path', required=True, type=str)
def download_file(uri: str, file_path: str) -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)
    try:
        data = biolib.lfs.get_file_data_from_large_file_system(lfs_uri=uri, file_path=file_path)
        with open(file_path, mode='wb') as file:
            file.write(data)
    except biolib_errors.BioLibError as error:
        print(f'An error occurred:\n{error.message}', file=sys.stderr)
        exit(1)


@lfs.command(help='Describe a Large File System')
@click.argument('uri', required=True)
@click.option('--json', is_flag=True, default=False, required=False, help='Format output as JSON')
def describe(uri: str, json: bool) -> None:
    describe_large_file_system(lfs_uri=uri, output_as_json=json)


@lfs.command(help='Prune LFS cache', hidden=True)
@click.option('--dry-run', type=click.BOOL, default=True, required=False)
def prune_cache(dry_run: bool) -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)
    prune_lfs_cache(dry_run)



import click

import biolib.runtime


@click.group(help='Commands available within a BioLib runtime')
def runtime() -> None:
    pass


@runtime.command(help='Set the name prefix of the main result')
@click.argument('result-prefix', required=True)
def set_main_result_prefix(result_prefix: str) -> None:
    biolib.runtime.set_main_result_prefix(result_prefix)



import logging
import click
from biolib.biolib_download_container import download_container_from_uri
from biolib.biolib_logging import logger, logger_no_user_data


@click.command(help='Push an application to BioLib', name='download-container')
@click.argument('uri')
def download_container(uri: str) -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)
    download_container_from_uri(uri)
